#ifndef _NVME_SPDK_CTRL_H
#define _NVME_SPDK_CTRL_H
#include <stdio.h>
#include <stdint.h>
#include <pthread.h>
#include "nvme_emu_ep.h"
#include "queue.h"

struct nvme_spdk_namespace;
#define NVME_SPDK_NAMESPACE_NAME_MAXLEN 16

#define MAX_SUBSYS_NUM        NVME_MAX_SUBSYSTEMS
#define MAX_CTRL_NUM          NVME_SUBSYS_MAX_CONTROLLERS
/* TODO: query num_queues from device cap */
#define NVME_CTRL_NUM_SQ_MAX  32
/* TODO: query SQ size from nvme_reg_cap_mqes */
#define MAX_SQ_SIZE           64

/* TODO: query mpsmax from nvme_reg_cap_mpsmax */
#define NVME_CTRL_PAGE_SHIFT  16
#define NVME_CTRL_PAGE_SIZE   (1 << NVME_CTRL_PAGE_SHIFT)
#define MAX_REQ_DATA          (NVME_CTRL_PAGE_SIZE * (1 << NVME_CRTL_MDTS_MAX))

typedef struct nvme_spdk_zcopy_ctx {
    void *fake_addr_table;
    size_t fake_addr_table_size;
    uintptr_t *request_table;
    struct snap_context *sctx;
    LIST_ENTRY(nvme_spdk_zcopy_ctx) entry;
} nvme_spdk_zcopy_ctx_t;

typedef struct nvme_spdk_ctrl {
    char name[NVME_EMU_NAME_MAXLEN];
    nvme_emu_ep_t *ep;
    int cntlid;

    /* list of private namespaces */
    TAILQ_HEAD(, nvme_spdk_namespace) namespaces;
    pthread_spinlock_t namespaces_lock;
    bool expose_namespaces;

    TAILQ_ENTRY(nvme_spdk_ctrl) entry;

    bool zcopy;
    nvme_spdk_zcopy_ctx_t *zcopy_ctx;
} nvme_spdk_ctrl_t;

int nvme_spdk_create_ctrl(nvme_emu_ep_t *ep);
void nvme_spdk_delete_ctrl(nvme_spdk_ctrl_t *ctrl);
void nvme_spdk_stop_ctrl(nvme_emu_ep_t *ep);
int nvme_spdk_start_ctrl(nvme_emu_ep_t *ep);
#endif
