#ifndef _VIRTIO_BLK_CTRL_H
#define _VIRTIO_BLK_CTRL_H
#include <stdio.h>
#include <stdbool.h>

typedef void (*vf_change_cb_t)(void *ctx, int pf_id, int num_vfs);

enum virtio_blk_ctrl_bdev_type {
    VIRTIO_BLK_CTRL_BDEV_NONE,
    VIRTIO_BLK_CTRL_BDEV_SPDK,
};

struct virtio_spdk_bdev_io_ctx {
    struct spdk_io_channel *spdk_channel;
    struct spdk_thread *spdk_thread;
    struct spdk_thread *caller_thread;
    struct virtio_blk_ctrl *ctrl;
};

struct virtio_spdk_bdev_ctx {
    struct spdk_bdev *bdev;
    struct spdk_bdev_desc *desc;
    struct virtio_spdk_bdev_io_ctx *io_channels;
    size_t num_io_channels;
};

struct virtio_blk_ctrl {
    size_t nthreads;
    enum virtio_blk_ctrl_bdev_type bdev_type;
    pthread_mutex_t bdev_lock;
    void (*bdev_close_cb)(struct virtio_blk_ctrl *ctrl);
    int pf_id;
    struct snap_context *sctx;
    struct ibv_pd *pd;
    struct snap_virtio_blk_ctrl *sctrl;
    struct virtio_spdk_bdev_ctx bctx;
    char serial[VIRTIO_BLK_ID_BYTES];
    void *vf_change_cb_arg;
    vf_change_cb_t vf_change_cb;
    bool zcopy;
};

struct virtio_blk_ctrl_init_attr {
    const char *emu_manager_name;
    int pf_id;
    int vf_id;
    const char *bdev_type;
    const char *bdev_name;
    uint32_t nthreads;
    uint32_t num_queues;
    uint32_t queue_depth;
    uint16_t size_max;
    uint16_t seg_max;
    const char *serial;
    void *vf_change_cb_arg;
    vf_change_cb_t vf_change_cb;
    bool force_in_order;
    bool suspended;
    bool recover;
};

struct virtio_blk_ctrl_io_stat
{
    uint64_t read_ios;
    uint64_t completed_read_ios;
    uint64_t completed_unord_r_ios;
    uint64_t write_ios;
    uint64_t completed_write_ios;
    uint64_t completed_unord_w_ios;
    uint64_t flush_ios;
    uint64_t completed_flush_ios;
    uint64_t completed_unord_f_ios;
    uint64_t err_read_os;
    uint64_t err_write_os;
    uint64_t err_flush_os;
};

int virtio_blk_ctrl_progress(void *ctrl);
int virtio_blk_ctrl_progress_all_io(void *ctrl);
int virtio_blk_ctrl_progress_io(void *arg, int thread_id);
void virtio_blk_ctrl_suspend(void *ctrl);
bool virtio_blk_ctrl_is_suspended(void *ctrl);
int virtio_blk_ctrl_resume(void *arg);
int virtio_blk_ctrl_state_save(void *ctrl, char *file);
int virtio_blk_ctrl_state_restore(void *ctrl, char *file);
int virtio_blk_ctrl_lm_enable(void *ctrl, char *lm_channel_name);
void virtio_blk_ctrl_lm_disable(void *ctrl);

int virtio_blk_ctrl_bdev_detach(void *ctrl_arg);
int virtio_blk_ctrl_bdev_attach(void *ctrl_arg, const char *bdev_type,
                                const char *bdev_name, uint16_t size_max,
                                uint16_t seg_max);
int virtio_blk_ctrl_get_debugstat(void *ctrl_arg,
                        struct snap_virtio_ctrl_debugstat *ctrl_debugstat);
struct virtio_blk_ctrl *
virtio_blk_ctrl_init(const struct virtio_blk_ctrl_init_attr *attr);
void virtio_blk_ctrl_destroy(void *ctrl);

void virtio_blk_ctrl_get_stat(struct virtio_blk_ctrl *ctrl, struct virtio_blk_ctrl_io_stat *stat);

#endif
